using System;
using System.Text;
using System.IO;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class DecoderTest {

    [SetUp]
    public void Setup()
    {
      // nothing to do
    }

    [TearDown]
    public void TearDown()
    {
      // nothing to do
    }

    [Test]
    public void TestDecodeContentTypeNotExist()
    {
      var message = new MemoryStream(Encoding.ASCII.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Subject: transfer encoding not exist\r\n" + 
        "\r\n" + 
        "decode as text/plain; charset=us-ascii"));

      var mime = MimeMessage.Create(message, true);

      Assert.IsNotNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);
      Assert.AreEqual("decode as text/plain; charset=us-ascii", mime.Body);
    }

    [Test]
    public void TestDecodeContentTypeNonText()
    {
      var message = new MemoryStream(Encoding.ASCII.GetBytes(
        "MIME-Version: 1.0\r\n" +
        "Content-Type: application/octet-stream\r\n" +
        "Subject: transfer encoding not exist\r\n" + 
        "\r\n" + 
        "decode as binary stream"));

      var mime = MimeMessage.Create(message, true);

      Assert.IsNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);

      var text = Encoding.ASCII.GetBytes("decode as binary stream");

      Assert.AreEqual(text.Length, mime.BodyBytes.Length);

      for (var i = 0; i < text.Length; i++) {
        Assert.AreEqual(text[i], mime.BodyBytes[i], string.Format("{0} byte", i));
      }
    }

    [Test]
    public void TestDecodeTransferEncoding7BitMessage()
    {
      var message = new MemoryStream(Encoding.ASCII.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: text/plain; charset=us-ascii\r\n" +
        "Content-Transfer-Encoding: 7bit\r\n" +
        "Subject: 7bit transfer encoding\r\n" + 
        "\r\n" + 
        "7bit transfer encoding"));

      var mime = MimeMessage.Create(message, true);

      Assert.IsNotNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);

      Assert.AreEqual("7bit", mime.Headers["Content-Transfer-Encoding"][0]);
      Assert.AreEqual("7bit transfer encoding", mime.Body);
    }

    [Test]
    public void TestDecodeTransferEncoding8BitMessage()
    {
      var message = new MemoryStream(latin1Encoding.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: text/plain; charset=iso-8859-1\r\n" +
        "Content-Transfer-Encoding: 8bit\r\n" +
        "Subject: 8bit transfer encoding\\r\n" + 
        "\r\n" + 
        "8bit transfer encoding \u0080 \u00cf \u00ff\r\n"));

      var mime = MimeMessage.Create(message, true);

      Assert.IsNotNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);

      Assert.AreEqual("8bit", mime.Headers["Content-Transfer-Encoding"][0]);
      Assert.AreEqual("8bit transfer encoding \u0080 \u00cf \u00ff\r\n", mime.Body);
    }

    [Test]
    public void TestDecodeTransferEncodingBinaryMessage()
    {
      var message = new MemoryStream(0x200);

      var header = latin1Encoding.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: application/octet-stream\r\n" +
        "Content-Transfer-Encoding: binary\r\n" +
        "Subject: binary transfer encoding\r\n" + 
        "\r\n");

      message.Write(header, 0, header.Length);

      var bodyStream = new MemoryStream(); 

      for (var i = 0; i < 0x100; i++) {
        bodyStream.WriteByte((byte)i);
      }

      var body = bodyStream.ToArray();

      message.Write(body, 0, body.Length);

      var mime = MimeMessage.Create(new MemoryStream(message.ToArray()), true);

      Assert.IsNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);

      Assert.AreEqual("binary", mime.Headers["Content-Transfer-Encoding"][0]);
      Assert.AreEqual(0x100, mime.BodyBytes.Length);
      
      for (var i = 0; i < 0x100; i++) {
        Assert.AreEqual(body[i], mime.BodyBytes[i], "{0} byte", i);
      }
    }

    [Test]
    public void TestDecodeTransferEncodingBase64Message()
    {
      var message = new MemoryStream(Encoding.ASCII.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: text/plain; charset=shift_jis\r\n" +
        "Content-Transfer-Encoding: base64\r\n" +
        "Subject: base64 transfer encoding\r\n" + 
        "\r\n" + 
        "ir+OmmFiY4KpgsgxMjODSoNpDQqCUIJRglKCgYKCgoMNCg=="));

      var mime = MimeMessage.Create(message, true);

      Assert.IsNotNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);

      Assert.AreEqual("base64", mime.Headers["Content-Transfer-Encoding"][0]);
      Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.Body);
    }

    [Test]
    public void TestDecodeTransferEncodingQuotedPrintableMessage()
    {
      var message = new MemoryStream(Encoding.ASCII.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: text/plain; charset=euc-jp\r\n" +
        "Content-Transfer-Encoding: quoted-printable\r\n" +
        "Subject: quoted printable transfer encoding\r\n" + 
        "\r\n" + 
        "=B4=C1=BB=FAabc=A4=AB=A4=CA123=A5=AB=A5=CA=0D=0A=A3=B1=A3=B2=A3=B3=A3=E1=A3=\r\n" +
        "=E2=A3=E3=0D=0A\r\n"));

      var mime = MimeMessage.Create(message, true);

      Assert.IsNotNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);

      Assert.AreEqual("quoted-printable", mime.Headers["Content-Transfer-Encoding"][0]);
      Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.Body);
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestDecodeTransferEncodingUnsupported()
    {
      var message = new MemoryStream(Encoding.ASCII.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: text/plain; charset=euc-jp\r\n" +
        "Content-Transfer-Encoding: unsupported\r\n" +
        "Subject: unsupported transfer encoding\r\n" + 
        "\r\n" + 
        "unsupported transfer encoding\r\n"));

      MimeMessage.Create(message, true);
    }

    [Test]
    public void TestDecodeTransferEncodingNotExist()
    {
      var message = new MemoryStream(Encoding.ASCII.GetBytes(
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: text/plain; charset=us-ascii\r\n" +
        "Subject: transfer encoding not exist\r\n" + 
        "\r\n" + 
        "decode as text/plain; charset=us-ascii"));

      var mime = MimeMessage.Create(message, true);

      Assert.IsNotNull(mime.Body);
      Assert.IsNotNull(mime.BodyBytes);
      Assert.AreEqual("decode as text/plain; charset=us-ascii", mime.Body);
    }

    [Test]
    public void TestDecodeTextQEncoded()
    {
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?utf-8?Q?=E6=BC=A2=E5=AD=97abc=E3=81=8B=E3=81=AA123=E3=82=AB=E3=83=8A?="),
                      "utf8");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?utf-7?Q?+byJbVw-abc+MEswag-123+MKswyg-?="),
                      "utf7");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?iso-2022-jp?Q?=1B$B4A;z=1B(Babc=1B$B$+$J=1B(B123=1B$B%+%J=1B(B?="),
                      "iso-2022-jp");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?shift_jis?Q?=8A=BF=8E=9Aabc=82=A9=82=C8123=83J=83i?="),
                      "shift_jis");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?euc-jp?Q?=B4=C1=BB=FAabc=A4=AB=A4=CA123=A5=AB=A5=CA?="),
                      "euc-jp");
    }

    [Test]
    public void TestDecodeTextBEncoded()
    {
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?utf-8?B?5ryi5a2XYWJj44GL44GqMTIz44Kr44OK?="),
                      "utf8");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?utf-7?B?K2J5SmJWdy1hYmMrTUVzd2FnLTEyMytNS3N3eWct?="),
                      "utf7");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?iso-2022-jp?B?GyRCNEE7ehsoQmFiYxskQiQrJEobKEIxMjMbJEIlKyVKGyhC?="),
                      "iso-2022-jp");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?shift_jis?B?ir+OmmFiY4KpgsgxMjODSoNp?="),
                      "shift_jis");
      Assert.AreEqual("漢字abcかな123カナ",
                      Decoder.DecodeMimeEncodedString("=?euc-jp?B?tMG7+mFiY6SrpMoxMjOlq6XK?="),
                      "euc-jp");

      Assert.AreEqual("【Microsoft】欽ちゃん球団の片岡安祐美さんが Office にチャレンジ! ワクワクの春 開幕!",
                      Decoder.DecodeMimeEncodedString(
                                                      "=?iso-2022-jp?B?GyRCIVobKEJNaWNyb3NvZnQbJEIhWzZWJEEkYyRzNWVDRBsoQg==?=" + 
                                                      " =?iso-2022-jp?B?GyRCJE5KUjIsMEJNNEh+JDUkcyQsGyhCIE9mZmljZSAbJEIkSyVBGyhC?=" + 
                                                      " =?iso-2022-jp?B?GyRCJWMlbCVzJTgbKEIhIBskQiVvJS8lbyUvJE49VRsoQiA=?=" + 
                                                      " =?iso-2022-jp?B?GyRCMytLaxsoQiE=?="));

      Assert.AreEqual("santamartaさんの日記にコメントが登録されました",
                      Decoder.DecodeMimeEncodedString("=?ISO-2022-JP?B?c2FudGFtYXJ0YRskQiQ1JHMkTkZ8NS0kSyUzJWElcyVIJCxFUE8/GyhC?= =?ISO-2022-JP?B?GyRCJDUkbCReJDckPxsoQg==?="));

      Assert.AreEqual("初めまして。突然のメールに深くお",
                      Decoder.DecodeMimeEncodedString("=?iso-2022-jp?B?GyRCPWkkYSReJDckRiEjRk1BMyROJWEhPCVrJEs/PCQvJCobKEI=?="));

      Assert.AreEqual("ウイルスバスタークラブニュース3月号■3年分でプラス3カ月。大好評キャンペーンが間もなく終了/新生活スタート。登録内容の変更を！",
                      Decoder.DecodeMimeEncodedString("=?ISO-2022-JP?B?GyRCJSYlJCVrJTklUCU5JT8hPCUvJWklViVLJWUhPCU5GyhC?=3=?ISO-2022-JP?B?GyRCN24bKEI=?=" +
                                                      " =?ISO-2022-JP?B?GyRCOWYiIxsoQg==?=3=?ISO-2022-JP?B?GyRCRy9KLCRHGyhC?=" +
                                                      " =?ISO-2022-JP?B?GyRCJVclaSU5GyhC?=3=?ISO-2022-JP?B?GyRCJSs3biEjGyhC?=" +
                                                      " =?ISO-2022-JP?B?GyRCQmc5JUk+JS0lYyVzJVohPCVzJCw0ViRiJEokLz0qTjsbKEI=?=" +
                                                      " /=?ISO-2022-JP?B?GyRCPzdAODNoJTklPyE8JUghI0VQTz9GYk1GJE5KUTk5GyhC?=" +
                                                      " =?ISO-2022-JP?B?GyRCJHIhKhsoQg==?="));
    }

    private Encoding latin1Encoding = Encoding.GetEncoding("latin1");
  }
}
