using System;
using System.Text;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class ParserTest {

    [SetUp]
    public void Setup()
    {
      // nothing to do
    }

    [TearDown]
    public void TearDown()
    {
      // nothing to do
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestUnsupportedMimeVersion()
    {
      var message = 
        "MIME-Version: 1.1\r\n" +
        "Content-Type: text/plain; charset=\"iso-2022-jp\"\r\n" +
        "Content-Transfer-Encoding: quoted-printable\r\n" + 
        "\r\n";

      MimeMessage.Create(message, false);
    }

    [Test]
    public void TestMimeVersionNotExist()
    {
      var message =
        "Content-Type: text/plain\r\n" +
        "From: from@example.com\r\n" + 
        "To: to@example.com\r\n";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["Content-Type"].Length);
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"][0]);
      Assert.AreEqual(1, mime.Headers["From"].Length);
      Assert.AreEqual("from@example.com", mime.Headers["From"][0]);
      Assert.AreEqual(1, mime.Headers["To"].Length);
      Assert.AreEqual("to@example.com", mime.Headers["To"][0]);
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestMimeVersionNotSupported()
    {
      var message =
        "MIME-Version: 1.1\r\n";

      MimeMessage.Create(message, false);
    }

    [Test]
    // TODO: [ExpectedException(typeof(NotSupportedException))]
    public void TestMimeVersionInvalid()
    {
      var message =
        "MIME-Version: \"1.0 (comment)\"\r\n";

      MimeMessage.Create(message, false);
    }

    [Test]
    public void TestMimeVersionWithComment()
    {
      var message =
        "MIME-Version: 1.0 (comment)\r\n";

      MimeMessage.Create(message, false);
    }

    [Test]
    public void TestMimeVersionQuoted()
    {
      var message =
        "MIME-Version: \"1.0\"\r\n";

      MimeMessage.Create(message, false);
    }

    [Test]
    public void TestIgnoreInvalidHeader()
    {
      var message =
        "MIME-Version 1.1\r\n" +
        "MIME-Version: 1.0\r\n";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["MIME-Version"].Length);
      Assert.AreEqual("1.0", mime.Headers["MIME-Version"][0]);
    }

    [Test]
    public void TestHeaderDelimiting()
    {
      var message =
        "Content-Type\t\t\t:\t\t\ttext/plain\r\n" +
        "From:       from@example.com\r\n" + 
        "To:to@example.com\r\n" +
        "Subject\t  : \tsubject\r\n";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["Content-Type"].Length);
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"][0]);
      Assert.AreEqual(1, mime.Headers["From"].Length);
      Assert.AreEqual("from@example.com", mime.Headers["From"][0]);
      Assert.AreEqual(1, mime.Headers["To"].Length);
      Assert.AreEqual("to@example.com", mime.Headers["To"][0]);
      Assert.AreEqual(1, mime.Headers["Subject"].Length);
      Assert.AreEqual("subject", mime.Headers["Subject"][0]);
    }

    [Test]
    public void TestCaseInvaliantComparison()
    {
      var message =
        "content-type: text/plain\r\n" +
        "FROM: from@example.com\r\n" + 
        "tO: to@example.com\r\n";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["CONTENT-TYPE"].Length);
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"][0]);
      Assert.AreEqual(1, mime.Headers["from"].Length);
      Assert.AreEqual("from@example.com", mime.Headers["FROM"][0]);
      Assert.AreEqual(1, mime.Headers["to"].Length);
      Assert.AreEqual("to@example.com", mime.Headers["TO"][0]);
    }

    [Test]
    public void TestMultipleHeader()
    {
      var message = 
        "MIME-Version: 1.0\r\n" +
        "Received: from A by B for C\r\n" +
        "Received: from X by Y for Z\r\n" +
        "Received: from 1 by 2 for 3\r\n";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(3, mime.Headers["Received"].Length);
      Assert.AreEqual("from A by B for C", mime.Headers["Received"][0]);
      Assert.AreEqual("from X by Y for Z", mime.Headers["RECEIVED"][1]);
      Assert.AreEqual("from 1 by 2 for 3", mime.Headers["received"][2]);
    }

    [Test]
    public void TestMultilineHeader()
    {
      var message = 
        "MIME-Version: 1.0\r\n" +
        "Subject: line1\r\n" +
        " line2\r\n" +
        "\tline3\r\n" +
        "   \tline4\r\n";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["Subject"].Length);
      Assert.AreEqual("line1 line2 line3 line4", mime.Headers["Subject"][0]);
    }

    [Test]
    public void TestSinglePartTextBody()
    {
      var message =
        "Content-Type: text/plain\r\n" +
        "Subject: single part\r\n" + 
        "\r\n" +
        "single-part text body";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["Content-Type"].Length);
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"][0]);
      Assert.AreEqual(1, mime.Headers["Subject"].Length);
      Assert.AreEqual("single part", mime.Headers["Subject"][0]);
      Assert.AreEqual("single-part text body", mime.Body);
    }

    [Test]
    public void TestMultiPartTextBody()
    {
      var message =
        "Content-Type: multipart/alternative;\r\n" +
        "  boundary=\"==boundary==\"\r\n" +
        "Subject: multi part\r\n" +
        "\r\n" +
        "multipart message body\r\n" +
        "\r\n" +
        "--==boundary==\r\n" +
        "Content-Type: text/plain\r\n" +
        "\r\n" +
        "first part, text/plain\r\n" +
        "\r\n" +
        "--==boundary==\r\n" +
        "Content-Type: text/html\r\n" +
        "\r\n" +
        "<html><body><p>second part, text/html</p></body></html>\r\n" +
        "--==boundary==--\r\n";

      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["Content-Type"].Length);
      Assert.AreEqual("multipart/alternative; boundary=\"==boundary==\"",
                      mime.Headers["Content-Type"][0]);
      Assert.AreEqual(1, mime.Headers["Subject"].Length);
      Assert.AreEqual("multi part", mime.Headers["Subject"][0]);
      Assert.AreEqual("multipart message body\r\n\r\n", mime.Body);

      Assert.AreEqual(2, mime.SubParts.Count);

      Assert.IsNotNull(mime.SubParts[0].Body);
      Assert.IsNull(mime.SubParts[0].BodyBytes);
      Assert.AreEqual(1, mime.SubParts[0].Headers["Content-Type"].Length);
      Assert.AreEqual("text/plain", mime.SubParts[0].Headers["Content-Type"][0]);
      Assert.AreEqual("first part, text/plain\r\n\r\n", mime.SubParts[0].Body);

      Assert.IsNotNull(mime.SubParts[1].Body);
      Assert.IsNull(mime.SubParts[1].BodyBytes);
      Assert.AreEqual(1, mime.SubParts[1].Headers["Content-Type"].Length);
      Assert.AreEqual("text/html", mime.SubParts[1].Headers["Content-Type"][0]);
      Assert.AreEqual("<html><body><p>second part, text/html</p></body></html>\r\n",
                      mime.SubParts[1].Body);
    }

    [Test]
    public void TestMultiPartWithNoHeader()
    {
      var message = "Return-Path: <>\r\n" + 
        "Date: Fri, 21 Dec 2007 07:04:35 +0900\r\n" + 
        "MIME-Version: 1.0\r\n" + 
        "Content-Type: multipart/report; report-type=delivery-status;\r\n" + 
        "	boundary=\"lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx\"\r\n" + 
        "Subject: Warning: could not send message for past 4 hours\r\n" + 
        "Auto-Submitted: auto-generated (warning-timeout)\r\n" + 
        "\r\n" + 
        "This is a MIME-encapsulated message\r\n" + 
        "\r\n" + 
        "--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx\r\n" + 
        "\r\n" + 
        "    **********************************************\r\n" + 
        "    **      THIS IS A WARNING MESSAGE ONLY      **\r\n" + 
        "    **  YOU DO NOT NEED TO RESEND YOUR MESSAGE  **\r\n" + 
        "    **********************************************\r\n" + 
        "\r\n" + 
        "The original message was received at Fri, 21 Dec 2007 02:06:24 +0900\r\n" + 
        "from xxx.xxxxxx.xxx [127.0.0.1]\r\n" + 
        "\r\n" + 
        "   ----- Transcript of session follows -----\r\n" + 
        "451 xxx.xxxxxx.xxx: Name server timeout\r\n" + 
        "Warning: message still undelivered after 4 hours\r\n" + 
        "Will keep trying until message is 5 days old\r\n" + 
        "451 xxx.xxxxxx.xxx: Name server timeout\r\n" + 
        "\r\n" + 
        "--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx\r\n" + 
        "Content-Type: message/delivery-status\r\n" + 
        "\r\n" + 
        "Reporting-MTA: dns; xxx.xxxxxx.xxx\r\n" + 
        "Arrival-Date: Fri, 21 Dec 2007 02:06:24 +0900\r\n" + 
        "\r\n" + 
        "Final-Recipient: RFC822; root@xxx.xxxxxx.xxx\r\n" + 
        "X-Actual-Recipient: RFC822; xxxxxx@xxx.xxxxxx.xxx\r\n" + 
        "Action: delayed\r\n" + 
        "Status: 4.4.3\r\n" + 
        "Last-Attempt-Date: Fri, 21 Dec 2007 07:04:35 +0900\r\n" + 
        "Will-Retry-Until: Wed, 26 Dec 2007 02:06:24 +0900\r\n" + 
        "\r\n" + 
        "--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx\r\n" + 
        "Content-Type: message/rfc822\r\n" + 
        "\r\n" + 
        "Return-Path: <root@xxx.xxxxxx.xxx>\r\n" + 
        "Date: Fri, 21 Dec 2007 02:02:15 +0900\r\n" + 
        "Content-Type: text/plain; charset=UTF-8\r\n" + 
        "Auto-Submitted: auto-generated\r\n" + 
        "X-Cron-Env: <MAILTO=root,postmaster,webmaster,clamav>\r\n" + 
        "X-Cron-Env: <SHELL=/bin/sh>\r\n" + 
        "X-Cron-Env: <HOME=/root>\r\n" + 
        "X-Cron-Env: <PATH=/usr/bin:/bin>\r\n" + 
        "X-Cron-Env: <LOGNAME=root>\r\n" + 
        "X-Cron-Env: <USER=root>\r\n" + 
        "\r\n" + 
        "ERROR: Can't query current.cvd.clamav.net\r\n" + 
        "ERROR: Can't get information about database.clamav.net: Temporary DNS error\r\n" + 
        "ERROR: No servers could be reached. Giving up\r\n" + 
        "ERROR: Can't query current.cvd.clamav.net\r\n" + 
        "ERROR: Can't get information about database.clamav.net: Temporary DNS error\r\n" + 
        "ERROR: No servers could be reached. Giving up\r\n" + 
        "ERROR: Can't query current.cvd.clamav.net\r\n" + 
        "ERROR: Can't get information about database.clamav.net: Temporary DNS error\r\n" + 
        "ERROR: No servers could be reached. Giving up\r\n" + 
        "ERROR: Update failed. Your network may be down or none of the mirrors listed in freshclam.conf is working.\r\n" + 
        "\r\n" + 
        "--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx--\r\n" + 
        "\r\n";
      var mime = MimeMessage.Create(message, false);

      Assert.IsNotNull(mime.Body);
      Assert.IsNull(mime.BodyBytes);

      Assert.AreEqual(1, mime.Headers["Content-Type"].Length);
      Assert.AreEqual("multipart/report; report-type=delivery-status; boundary=\"lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx\"",
                      mime.Headers["Content-Type"][0]);

      Assert.AreEqual(3, mime.SubParts.Count);

      Assert.AreEqual(0, mime.SubParts[0].Headers.Count);
      Assert.IsTrue(mime.SubParts[0].Body.StartsWith("    ***"));

      Assert.AreEqual("message/delivery-status", mime.SubParts[1].Headers["Content-Type"][0]);
      Assert.AreEqual("message/rfc822", mime.SubParts[2].Headers["Content-Type"][0]);
    }
  }
}